package be.cylab.java.wowa.training;

import com.owlike.genson.GenericType;
import com.owlike.genson.Genson;
import info.debatty.java.aggregation.WOWA;
import org.apache.commons.lang3.ArrayUtils;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

/**
 * Class Utils. Contain tools for Trainer algorithm
 */
final class Utils {

    private Utils() {

    }

    /**
     * Normalize a weights vector.
     *
     * @param weights
     * @return
     */
    static List<Double> normalizeWeights(final List<Double> weights) {
        Double sum_weight = Utils.sumListElements(weights);
        if (sum_weight == 0.0) {
            throw new IllegalArgumentException(
                    "Sum of weights must be different of 0");
        }
        List<Double> weights_normalized = new ArrayList<>();
        for (int i = 0; i < weights.size(); i++) {
            weights_normalized.add(weights.get(i) / sum_weight);
        }
        return weights_normalized;
    }

    /**
     * Find the max distance in a List of SolutionDistance.
     *
     * @param solutions
     * @return
     */
    static double findMaxDistance(
            final List<AbstractSolution> solutions) {
        double max = Double.NEGATIVE_INFINITY;
        for (AbstractSolution solution : solutions) {
            if (solution.getFitnessScore() > max) {
                max = solution.getFitnessScore();
            }
        }
        return max;
    }

    /**
     * Find the minimum distance in a List of SolutionDistance.
     *
     * @param solutions
     * @return
     */
    static double findMinDistance(
            final List<AbstractSolution> solutions) {
        double min = Double.POSITIVE_INFINITY;
        for (AbstractSolution solution : solutions) {
            if (solution.getFitnessScore() < min) {
                min = solution.getFitnessScore();
            }
        }
        return min;
    }

    /**
     * Compute the sum of all distance in a List of SolutionDistance.
     *
     * @param solutions
     * @return
     */
    static double sumTotalDistance(
            final List<AbstractSolution> solutions) {
        double sum = 0;
        for (AbstractSolution solution : solutions) {
            sum += solution.getFitnessScore();
        }
        return sum;
    }

    /**
     * Compute the sum of all array's element.
     *
     * @param array
     * @return
     */
    static Double sumListElements(final List<Double> array) {
        Double sum = 0.0;
        for (Double weight : array) {
            sum += weight;
        }
        return sum;
    }

    /**
     * Generate a random integer in a range.
     *
     * @param min
     * @param max
     * @return
     */
    static int randomInteger(final int min, final int max) {
        if (min > max) {
            throw new IllegalArgumentException("Max must be greater then min");
        } else if (min == max) {
            return min;
        }
        Random rnd = new Random();
        return rnd.nextInt((max - min) + 1) + min;
    }

    /**
     * Import a json file.
     *
     * @param filename
     * @return
     */
    static List<List<Double>> convertJsonToDataForTrainer(
            final String filename) {
        Genson genson = new Genson();
        String data_json = Utils.readFileToString(filename);
        List<List<Double>> data = genson.deserialize(
                data_json,
                new GenericType<List<List<Double>>>() {
                });
        return data;
    }

    /**
     * Import result file.
     *
     * @param filename
     * @return
     */
    static List<Double> convertJsonToExpectedForTrainer(
            final String filename) {

        Genson genson = new Genson();
        String expected_json = Utils.readFileToString(filename);
        List<Double> expected = genson.deserialize(expected_json,
                new GenericType<List<Double>>() {
                });
        return expected;
    }

    /**
     * Method to read file and convert into a String.
     *
     * @param filename
     * @return
     */
    private static String readFileToString(final String filename) {
        String json = null;
        try {
            json = new String(Files.readAllBytes(Paths.get(filename)),
                    StandardCharsets.UTF_8);
        } catch (IOException e) {
            e.printStackTrace();
        }
        return json;
    }


    /**
     * @param expected
     * @return
     */
    static boolean[] convertExpectedToBooleanArrayTrueAlert(
            final List<Double> expected) {
        boolean[] true_alert = new boolean[expected.size()];
        for (int i = 0; i < expected.size(); i++) {
            if (expected.get(i) == 1) {
                true_alert[i] = true;
            } else if (expected.get(i) == 0) {
                true_alert[i] = false;
            } else {
                throw new IllegalStateException(
                        "Expected must be equal to 0 or 1 to compute ROC");
            }
        }
        return true_alert;
    }

    /**
     * Function to convert a List of doubles to an Array of doubles.
     *
     * @param elements
     * @return
     */
    static double[] convertListDoubleToArrayDouble(
            final List<Double> elements) {
        Double[] w = new Double[elements.size()];
        w = elements.toArray(w);
        double[] ww = ArrayUtils.toPrimitive(w);
        return ww;
    }

    /**
     * @param size
     * @return
     */
    static List<Double> initializeListWithZeroValues(
            final int size) {
        List<Double> list = new ArrayList<>();
        for (int i = 0; i < size; i++) {
            list.add(0.0);
        }
        return list;
    }


    /**
     * Correct weights vectors if it is null.
     * With QUASI_RANDOM initialization, it is possible to obtain a full null
     * weight vector after mating. The method check and correct that
     *
     * @param child AbstractSolution
     */
    static void checkAndCorrectNullWeightVector(final AbstractSolution child) {
        if (Utils.sumListElements(child.getWeightsW()) == 0) {
            int index = Utils.randomInteger(0, child.getWeightsW().size() - 1);
            List<Double> w = child.getWeightsW();
            w.set(index, 1.0);
            child.setWeightsW(w);
        }
        if (Utils.sumListElements(child.getWeightsP()) == 0) {
            int index = Utils.randomInteger(0, child.getWeightsP().size() - 1);
            List<Double> p = child.getWeightsP();
            p.set(index, 1.0);
            child.setWeightsP(p);
        }
    }

    static void computeStatisticalInformation(
            final List<List<Double>> data,
            final List<Double> expected,
            final int nbr_values,
            final String filename,
            final AbstractSolution solution) {
        int wowa_true_positive_counter = 0;
        int wowa_true_negative_counter = 0;
        int average_true_positive_counter = 0;
        int average_true_negative_counter = 0;

        int wowa_false_positive_counter = 0;
        int wowa_false_negative_counter = 0;
        int average_false_positive_counter = 0;
        int average_false_negative_counter = 0;

        double iteration = 1 / (double) nbr_values;


        for (double trigger_value = 0.00;
             trigger_value < 1; trigger_value += iteration) {
            for (int j = 0; j < data.size(); j++) {
                WOWA wowa = new WOWA(Utils.convertListDoubleToArrayDouble(
                        solution.weights_w),
                        Utils.convertListDoubleToArrayDouble(
                                solution.weights_p));
                double wowa_result = wowa.aggregate(
                        Utils.convertListDoubleToArrayDouble(data.get(j)));
                double average = Utils.sumListElements(data.get(j))
                        / data.get(j).size();
                if (expected.get(j) == 1 && average > trigger_value && wowa_result > trigger_value) {
                    average_true_positive_counter++;
                    wowa_true_positive_counter++;
                } else if (expected.get(j) == 1 && average < trigger_value && wowa_result > trigger_value) {
                    wowa_true_positive_counter++;
                    average_false_negative_counter++;
                } else if (expected.get(j) == 1 && average > trigger_value && wowa_result < trigger_value) {
                    average_true_positive_counter++;
                    wowa_false_negative_counter++;
                } else if (expected.get(j) == 1 && average < trigger_value && wowa_result < trigger_value) {
                    average_false_negative_counter++;
                    wowa_false_negative_counter++;
                } else if (expected.get(j) == 0 && average < trigger_value && wowa_result < trigger_value) {
                    average_true_negative_counter++;
                    wowa_true_negative_counter++;
                } else if (expected.get(j) == 0 && average < trigger_value && wowa_result > trigger_value) {
                    average_true_negative_counter++;
                    wowa_false_positive_counter++;
                } else if (expected.get(j) == 0 && average > trigger_value && wowa_result < trigger_value) {
                    wowa_true_negative_counter++;
                    average_false_positive_counter++;
                } else if (expected.get(j) == 0 && average > trigger_value && wowa_result > trigger_value) {
                    average_false_positive_counter++;
                    wowa_false_positive_counter++;
                }
            }

            //Positive conditions.
            int wowa_positive_condition = wowa_true_positive_counter + wowa_false_negative_counter;
            int average_positive_condition = average_true_positive_counter + average_false_negative_counter;

            //Negative conditions.
            int wowa_negative_condition = wowa_true_negative_counter + wowa_false_positive_counter;
            int average_negative_condition = average_true_negative_counter + average_false_positive_counter;

            //True positive rate.
            double wowa_true_positive_rate = wowa_true_positive_counter / wowa_positive_condition;
            double average_true_positive_rate = average_true_positive_counter / average_positive_condition;

            //True negative rate.
            double wowa_true_negative_rate = wowa_true_negative_counter / wowa_negative_condition;
            double average_true_negative_rate = average_true_negative_counter / average_negative_condition;

            //False negative rate.
            double wowa_false_negative_rate = 1 - wowa_true_positive_rate;
            double average_false_negative_rate = 1 - average_true_positive_rate;

            //Flase positive rate.
            double wowa_false_positive_rate = 1 - wowa_true_negative_rate;
            double average_false_positive_rate = 1 - average_true_negative_rate;

            //Precision (or positive prediction value).
            double wowa_precision = wowa_true_positive_counter / (double) (wowa_true_positive_counter + wowa_false_positive_counter);
            double average_precision = average_true_positive_counter / (double) (average_true_positive_counter + average_false_positive_counter);

            //Negative prediction value.
            double wowa_negative_prediction_value = wowa_true_negative_counter / (double) (wowa_true_negative_counter + wowa_false_negative_counter);
            double average_negative_prediction_value = average_true_negative_counter / (double) (average_true_negative_counter + average_false_negative_counter);

            //False discovery rate.
            double wowa_false_discovery_rate = 1 - wowa_precision;
            double average_false_discovery_rate = 1 - average_precision;

            //False omission rate.
            double wowa_false_omission_rate = 1 - wowa_negative_prediction_value;
            double average_false_omission_rate = 1 - average_negative_prediction_value;

            //Accuracy.
            double wowa_accuracy = (wowa_true_positive_counter + wowa_true_negative_counter) / expected.size();
            double average_accuracy = (average_true_positive_counter + average_true_negative_counter) / expected.size();

            //F1 score.
            double wowa_f_1 = 2 * (wowa_precision * wowa_true_positive_rate) / (wowa_precision + wowa_true_positive_rate);
            double average_f_1 = 2 * (average_precision * average_true_positive_rate) / (average_precision + average_true_positive_rate);

            try (FileWriter fw = new FileWriter(filename, true);
                 BufferedWriter bw = new BufferedWriter(fw);
                 PrintWriter writer = new PrintWriter(bw)) {

                writer.println("___________________________________________");
                writer.println("|TRIGGER VALUE       :" + trigger_value);
                writer.println("|_________________________________________|");
                writer.println("WOWA");

                writer.println("Wowa true positive : " + wowa_true_positive_counter);
                writer.println("Wowa true negative : " + wowa_true_negative_counter);
                writer.println("Wowa false positive : " + wowa_false_positive_counter);
                writer.println("Wowa false negative : " + wowa_false_negative_counter);

                writer.println("Wowa true positive rate : " + wowa_true_positive_rate);
                writer.println("Wowa true negative rate : " + wowa_true_negative_rate);
                writer.println("Wowa false positive rate : " + wowa_false_positive_rate);
                writer.println("Wowa false negative rate : " + wowa_false_negative_rate);

                writer.println("Wowa precision : " + wowa_precision);
                writer.println("Wowa negative predictive value : " + wowa_negative_prediction_value);
                writer.println("Wowa false discovery rate : " + wowa_false_discovery_rate);
                writer.println("Wowa false omission rate : " + wowa_false_omission_rate);
                writer.println("Wowa accuracy : " + wowa_accuracy);
                writer.println("Wowa F1 score : " + wowa_f_1);

                writer.println("AVERAGE\n");
                writer.println("Average true positive : " + average_true_positive_counter);
                writer.println("Average true negative : " + average_true_negative_counter);
                writer.println("Average false positive : " + average_false_positive_counter);
                writer.println("Average false negative : " + average_false_negative_counter);

                writer.println("Average true positive rate : " + average_true_positive_rate);
                writer.println("Average true negative rate : " + average_true_negative_rate);
                writer.println("Average false positive rate : " + average_false_positive_rate);
                writer.println("Average false negative rate : " + average_false_negative_rate);

                writer.println("Average precision : " + average_precision);
                writer.println("Average negative predictive value : " + average_negative_prediction_value);
                writer.println("Average false discovery rate : " + average_false_discovery_rate);
                writer.println("Average false omission rate : " + average_false_omission_rate);
                writer.println("Average accuracy : " + average_accuracy);
                writer.println("Average F1 score : " + average_f_1);

                wowa_true_positive_counter = 0;
                wowa_true_negative_counter = 0;
                average_true_positive_counter = 0;
                average_true_negative_counter = 0;

                wowa_false_positive_counter = 0;
                wowa_false_negative_counter = 0;
                average_false_positive_counter = 0;
                average_false_negative_counter = 0;

            } catch (FileNotFoundException e) {
                e.printStackTrace();
            } catch (UnsupportedEncodingException e) {
                e.printStackTrace();
            } catch (IOException e) {
                e.printStackTrace();
            }

        }

    }

}
